function LidarPost(lidarclass, classfile, classprobfile, ndemfile, slopefile, roughfile, bridgefile)

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%NOTE TO REVIEWERS!

%Minor changes have been made to this script since it has been used for the
%work documented in the manuscript but the essential mechanism is still the
%same.

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%SUBFUNCTIONS (these are here just to read and write ENVI files in Matlab)
%Code developed by Felix Totir (2010)
function info = envihdrread(hdrfile)
    
fid = fopen(hdrfile);
while true
    line = fgetl(fid);
    if line == -1
        break
    else
        eqsn = findstr(line,'=');
        if ~isempty(eqsn)
            param = strtrim(line(1:eqsn-1));
            param(findstr(param,' ')) = '_';
            value = strtrim(line(eqsn+1:end));
            if isempty(str2num(value))
                if ~isempty(findstr(value,'{')) && isempty(findstr(value,'}'))
                    while isempty(findstr(value,'}'))
                        line = fgetl(fid);
                        value = [value,strtrim(line)];
                    end
                end
                eval(['info.',param,' = ''',value,''';'])
            else
                eval(['info.',param,' = ',value,';'])
            end
        end
    end
end
fclose(fid);
if isfield(info,'map_info')
    line = info.map_info;
    line(line == '{' | line == '}') = [];
    line=textscan(line,'%s',',');
    line=line{:};
    line=strtrim(line);
    info.map_info = [];
    info.map_info.projection = line{1};
    info.map_info.image_coords = [str2num(line{2}),str2num(line{3})];
    info.map_info.mapx = str2num(line{4});
    info.map_info.mapy = str2num(line{5});
    info.map_info.dx  = str2num(line{6});
    info.map_info.dy  = str2num(line{7});
    if length(line) == 9
        info.map_info.datum  = line{8};
        info.map_info.units  = line{9}(7:end);
    elseif length(line) == 11
        info.map_info.zone  = str2num(line{8});
        info.map_info.hemi  = line{9};
        info.map_info.datum  = line{10};
        info.map_info.units  = line{11}(7:end);
    end
end

if isfield(info,'pixel_size')
    line = info.pixel_size;
    line(line == '{' | line == '}') = [];
    line=textscan(line,'%s',',');
    line=line{:};
    line=strtrim(line);
    info.pixel_size = [];
    info.pixel_size.x = str2num(line{1});
    info.pixel_size.y = str2num(line{2});
    info.pixel_size.units = line{3}(7:end);
end
end
%--------------------------------------------------------------------------
%Code developed by Felix Totir (2010)
function D = enviread(filename, headername)
    
D_info = envihdrread(headername);

%Set binary format parameters
switch D_info.byte_order
    case {0}
        machine = 'ieee-le';
    case {1}
        machine = 'ieee-be';
    otherwise
        machine = 'n';
end
iscx=false; %if it is complex
switch D_info.data_type
    case {1}
        format = 'uint8';
    case {2}
        format= 'int16';
    case{3}
        format= 'int32';
    case {4}
        format= 'single';
    case {5}
        format= 'double';
    case {6}
        iscx=true;
        format= 'single';
    case {9}
        iscx=true;
        format= 'double';
    case {12}
        format= 'uint16';
    case {13}
        format= 'uint32';
    case {14}
        format= 'int64';
    case {15}
        format= 'uint64';
    otherwise
        error(['File type number: ',num2str(dtype),' not supported']);
end
D = multibandread(filename,[D_info.lines, D_info.samples, D_info.bands], format, 0, D_info.interleave, machine);
end
%--------------------------------------------------------------------------
%Code developed by Felix Totir (2010)
function envihdrwrite(info,hdrfile)

params=fieldnames(info);
params2 = fieldnames(info.map_info);
fid = fopen(hdrfile,'w');
fprintf(fid,'%s\n','ENVI');
for idx=1:length(params)
    param = params{idx};
    value = getfield(info,param);
    param(findstr(param,'_')) = ' ';
    if isstruct(value) == true
        line=[param ' = ' '{ '];
        for idy=1:length(params2)
            param2 = params2{idy};
            value2 = getfield(info.map_info,param2);
            param2(findstr(param2,'_')) = ' ';
            if numel(value2) == 2
                line = [line num2str(value2(1)) ',' num2str(value2(2))];
            else
                line=[line num2str(value2)];
            end
            if idy < length(params2) && line(end) ~= ','
                line = [line ','];
            end
        end
        line = [line ' } '];
    else
        line=[param,' = ',num2str(value)];
    end 
    fprintf(fid,'%s\r\n',line);
end
fclose(fid);
end
%--------------------------------------------------------------------------
%Code developed by Felix Totir (2010)
function envidatawrite(D,datafile,info)

if nargin < 2
    error('You must specify the information about data (input of envihdrwrite)');
end

% Set binary format parameters
switch info.byte_order
    case {0}
        machine = 'ieee-le';
    case {1}
        machine = 'ieee-be';
    otherwise
        machine = 'n';
end
iscx=false;
switch info.data_type
    case {1}
        format = 'uint8';
    case {2}
        format= 'int16';
    case{3}
        format= 'int32';
    case {4}
        format= 'single';
    case {5}
        format= 'double';
    case {6}
        iscx=true;
        format= 'single';
    case {9}
        iscx=true;
        format= 'double';
    case {12}
        format= 'uint16';
    case {13}
        format= 'uint32';
    case {14}
        format= 'int64';
    case {15}
        format= 'uint64';
    otherwise
        error(['File type number: ',num2str(dtype),' not supported']);
end
switch lower(info.interleave)
    case {'bsq'}
        D = permute(D,[2,1,3]);
    case {'bil'}
        D = permute(D,[2,3,1]);
    case {'bip'}
        D = permute(D,[3,2,1]);
end
D=D(:);
if iscx
    D = [real(D), imag(D)];
    D=D.';
    D=D(:);
end
fid=fopen(datafile,'w');
if isfield(info, 'header_offset') == 0
    info.header_offset = 0;
end
fwrite(fid,zeros(info.header_offset,1),'uint8',0,machine);
fwrite(fid,D,format,0,machine);
fclose(fid);
end
%--------------------------------------------------------------------------
%Code developed by Felix Totir (2010)
function enviwrite(D,info,datafile,hdrfile)

if nargin < 3
    error('You must specify at least three inputs');
elseif nargin < 4
    hdrfile=[deblank(datafile),'.hdr'];
end
envihdrwrite(info,hdrfile);
envidatawrite(D,datafile,info);
end
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%MAIN FUNCTION (where the actual LiDAR correction of SVC happens)

tic;

%define model parameters
ndemtres = 0.5;
slopetres = 10;
roughtres = 20;

%process input
classhead = [classfile '.hdr'];
classprobhead = [classprobfile '.hdr'];
ndemhead = [ndemfile '.hdr'];
slopehead = [slopefile '.hdr'];
roughhead = [roughfile '.hdr'];
bridgehead = [bridgefile '.hdr'];

classimage = enviread(classfile, classhead);
classprob = enviread(classprobfile, classprobhead);
ndem = enviread(ndemfile, ndemhead);
slope = enviread(slopefile, slopehead);
rough = enviread(roughfile, roughhead);
bridge = enviread(bridgefile, bridgehead);

class_info = envihdrread(classhead);
samples = class_info.samples;
lines = class_info.lines;
classes = class_info.classes;

%save intitial class probability image
%classprob_in = max(classprob,[],3);
%classprob_info = envihdrread(classprobhead);
%classprob_info.bands = 1;
%classprob_info.band_names = '{ class probability }';
%classprob_infile = [classprobfile '_initial'];
%classprob_inhead = [classprob_infile '.hdr'];
%enviwrite(classprob_in, classprob_info, classprob_infile, classprob_inhead)
%clear classprob_in

%initialize additional model variables
change = 1;
counter = 1;

%create observed LiDAR classes
ndem(ndem < ndemtres) = 0;
ndem(ndem >= ndemtres) = 1;
ndem(bridge == 1) = 0;
slope(slope < slopetres) = 0;
slope(slope >= slopetres) = 1;
rough(rough < roughtres) = 0;
rough(rough >= roughtres) = 1;

%create look-up table from predefined lidarclass file
LUTndem = [lidarclass(1,:) classes+1:256];
LUTslope = [lidarclass(2,:), classes+1:256];

%main loop
while change == 1
    %detect possible conflicts between class image and lidar classes
    conflict = zeros(lines, samples);
    ndemclass = intlut(uint8(classimage),uint8(LUTndem));
    ndemclass(ndemclass == 2) = ndem(ndemclass == 2);
    slopeclass = intlut(uint8(classimage),uint8(LUTslope));
    slopeclass(slopeclass == 2 | rough == 1) = slope(slopeclass == 2 | rough == 1);
    conflict(ndem ~= ndemclass) = 1;
    conflict(slope ~= slopeclass) = 1;
    
    %in case of conflict, set class probabilities of corresponding pixels
    %to zero, compute new class image and repeat loop
    [I,J] = find(conflict == 1);
	for i=1:numel(I)
        if classimage(I(i),J(i)) > 0
            classprob(I(i),J(i),classimage(I(i),J(i))) = 0;
        end
    end
    if any(conflict) == 0
        change = 0;
    end
    counter = counter + 1;
    if counter == classes
        change = 0;
    end
    [Y,classimage] = max(classprob,[],3);
    clear Y
    ['iteration ' num2str(counter)]
    toc
end

['total iterations =' num2str(counter)]

%save final class
classfile = [classfile '_LC'];
classhead = [classfile '.hdr'];
enviwrite(classimage, class_info, classfile, classhead)

%save final SVM probability image
%classprob_out = max(classprob,[],3);
%classprob_outfile = [classprobfile '_final'];
%classprob_outhead = [classprob_infile '.hdr'];
%enviwrite(classprob_out, classprob_info, classprob_outfile, classprob_outhead)
            
end

%-----------------------------------------------------------------------------
%-----------------------------------------------------------------------------
%REFERENCES

%Felix Totir, 2010, ENVI reader/writer, downloaded from:
%http://www.mathworks.com/matlabcentral/fileexchange/27172-envi-file-reader-writer

%-----------------------------------------------------------------------------
%-----------------------------------------------------------------------------